"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.run = void 0;
const path = require("path");
const fs = require("fs");
const nodegit_1 = require("nodegit");
const rules_1 = require("../lib/rules");
const check = async (ctx, entry, blobString, ruleType) => {
    var _a;
    const rule = rules_1.rules[ruleType];
    if (!rule || !ctx.currentRepo)
        return;
    const matches = blobString.match(rule.pattern);
    if (matches) {
        ctx.log.debug(`[!] found ${matches.length} matches for ${ruleType}`);
        const blame = await nodegit_1.Blame.file(ctx.currentRepo, entry.path());
        for (const matched of matches) {
            if (rule.antipattern !== undefined && rule.antipattern.test(matched)) {
                ctx.log.debug(`match excluded by antipattern rule ${rule.antipattern}`);
                continue;
            }
            let lineNumber = 0;
            for (const line of blobString.split("\n")) {
                lineNumber++;
                if (line.includes(matched)) {
                    const valuePart = ctx.redactValues ? "" : `found ${matched} `;
                    let blameFinding = undefined;
                    const blameHunk = (await blame).getHunkByLine(lineNumber);
                    const blameCommitId = blameHunk.finalCommitId();
                    try {
                        const blameCommit = await ctx.currentRepo.getCommit(blameCommitId);
                        blameFinding = {
                            commit: blameCommit.toString(),
                            author: {
                                email: blameCommit.committer().email(),
                            },
                            date: blameCommit.date(),
                        };
                    }
                    catch (e) {
                        ctx.log.warn(`tried to blame but the related commit ${blameCommitId} has been deleted`);
                    }
                    const finding = {
                        ruleType,
                        detail: `${valuePart}in ${entry.toString()} on line ${lineNumber}`,
                        blame: blameFinding,
                    };
                    (_a = ctx.result) === null || _a === void 0 ? void 0 : _a.findings.push(finding);
                    ctx.log.warn(`(${ctx.currentRepoName}) [${finding.ruleType}] ${finding.detail}` +
                        (finding.blame
                            ? ` (${finding.blame.commit} by ${finding.blame.author.email} at ${finding.blame.date.toISOString()})`
                            : ""));
                }
            }
        }
    }
};
const testEntry = async (ctx, entry) => {
    var _a;
    ctx.log.debug(`checking ${ctx.currentPath}/${entry.toString()}...`);
    let blob;
    try {
        blob = await entry.getBlob();
    }
    catch (e) {
        ctx.log.debug("no blob data");
        return;
    }
    try {
        const blobString = blob.toString();
        for (const ruleType in rules_1.rules) {
            const filterEntry = (_a = rules_1.rules[ruleType]) === null || _a === void 0 ? void 0 : _a.filterEntry;
            if (!filterEntry || filterEntry.test(entry.toString()))
                await check(ctx, entry, blobString, ruleType);
        }
    }
    catch (e) {
        ctx.log.error(e);
    }
};
const openRepo = async (ctx, dir) => {
    try {
        return await nodegit_1.Repository.open(path.resolve(dir));
    }
    catch (e) {
        ctx.log.error("failed to open repository");
        throw e;
    }
};
const latestCommit = async (ctx) => {
    try {
        return await ctx.currentRepo.getHeadCommit();
    }
    catch (e) {
        ctx.log.error("failed to get latest HEAD commit");
        throw e;
    }
};
const checkTreeEntries = async (ctx, entries) => {
    for (const entry of entries) {
        if (entry.isTree())
            await checkTreeEntries(ctx, (await entry.getTree()).entries());
        await testEntry(ctx, entry);
    }
};
const repoNameFromPath = (path) => {
    const pathParts = path.split("/");
    return pathParts[pathParts.length - 1];
};
const scanRepoDir = async (ctx, dir) => {
    ctx.log.info(`scanning ${dir}`);
    try {
        ctx.currentRepo = await openRepo(ctx, dir);
        ctx.currentPath = dir;
        ctx.currentRepoName = repoNameFromPath(dir);
        const commit = await latestCommit(ctx);
        const branchName = (await ctx.currentRepo.getCurrentBranch()).shorthand();
        ctx.log.info(`analysing files at commit ${commit.toString()} (${branchName})`);
        const tree = await commit.getTree();
        await checkTreeEntries(ctx, tree.entries());
    }
    catch (e) {
        ctx.log.warn("directory not scanned - check debug log for details");
        ctx.log.debug(e);
    }
};
exports.run = async (ctx) => {
    try {
        const resolvedDir = path.resolve(ctx.rootDirectory);
        // get a handle on the .git directory
        const gitDir = path.resolve(resolvedDir, ".git");
        ctx.log.info(`SCAN STARTED: ${resolvedDir}`);
        ctx.log.debug(`checking for ${gitDir}`);
        // initialise result
        ctx.result = { findings: [] };
        // is this a git repo?
        if (!fs.existsSync(gitDir)) {
            // not a git repo, iterate directories
            ctx.log.debug(`${resolvedDir} is not a git repo, digging...`);
            const files = await fs.promises.readdir(resolvedDir);
            for (const file of files) {
                const fullPath = path.resolve(resolvedDir, file);
                const stat = await fs.promises.stat(fullPath);
                if (stat.isDirectory())
                    await scanRepoDir(ctx, fullPath);
            }
        }
        else
            await scanRepoDir(ctx, resolvedDir);
    }
    catch (e) {
        ctx.log.error(e, "SCAN FAILED");
        throw e;
    }
    ctx.log.info(`SCAN COMPLETE (found ${ctx.result.findings.length} issues)`);
};
